@@ -1,7 +1,9 @@
 /* An "assembler" for PTX.
    Copyright (C) 2014, 2015 Mentor Graphics
-   Contributed by Nathan Sidwell <nathan@codesourcery.com>
-   Contributed by Bernd Schmidt <bernds@codesourcery.com>
+   Copyright (C) 2017 Red Hat
+   Copyright (C) 2017, 2021, 2022, 2023 Siemens
+   Copyright (C) 2020 SUSE
+   Copyright (C) 2024 BayLibre
 
    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
@@ -26,7 +28,6 @@
 #include <getopt.h>
 #include <stdlib.h>
 #include <stdio.h>
-#include <stdarg.h>
 #include <string.h>
 #include <sys/wait.h>
 #include <unistd.h>
@@ -41,6 +42,10 @@
 #include <hashtab.h>
 
 #include <list>
+#include <vector>
+#include <iostream>
+#include <fstream>
+#include <sstream>
 
 #include "version.h"
 
@@ -80,16 +85,10 @@ static bool verbose = false;
 
 static const char *outname = NULL;
 
-static void __attribute__ ((format (printf, 1, 2)))
-fatal_error (const char * cmsgid, ...)
+static void
+fatal_error (const std::string &error_string)
 {
-  va_list ap;
-
-  va_start (ap, cmsgid);
-  fprintf (stderr, "nvptx-as: ");
-  vfprintf (stderr, cmsgid, ap);
-  fprintf (stderr, "\n");
-  va_end (ap);
+  std::cerr << "nvptx-as: " << error_string << "\n";
 
   if (outname)
     unlink (outname);
@@ -149,26 +148,29 @@ hash_string_hash (const void *e_p)
    Takes ownership of STRING.  */
 
 static symbol *
-symbol_hash_lookup (htab_t symbol_table, char *string)
+symbol_hash_lookup (htab_t symbol_table, std::vector<symbol *> &symbols, char *string)
 {
-  void **e;
-  e = htab_find_slot_with_hash (symbol_table, string,
-                                (*htab_hash_string) (string),
-                                INSERT);
+  symbol **e;
+  e = (symbol **) htab_find_slot_with_hash (symbol_table, string,
+					    (*htab_hash_string) (string),
+					    INSERT);
   if (e == NULL)
     {
       free (string);
       return NULL;
     }
   if (*e == NULL)
-    *e = new symbol (string);
+    {
+      *e = new symbol (string);
+      symbols.push_back (*e);
+    }
   else
     free (string);
 
-  return (symbol *) *e;
+  return *e;
 }
 
-typedef enum Kind
+enum Kind
 {
   /* 0-ff used for single char tokens */
   K_symbol = 0x100, /* a symbol */
@@ -178,9 +180,9 @@ typedef enum Kind
   K_number,
   K_string,
   K_comment
-} Kind;
+};
 
-typedef struct Token
+struct Token
 {
   unsigned short kind : 12;
   unsigned short space : 1; /* preceded by space */
@@ -190,13 +192,13 @@ typedef struct Token
 
   /* Token itself */
   char const *ptr;
-} Token;
+};
 
-/* The preamble '.target' directive's argument.  */
-static char *preamble_target_arg;
+#define is_keyword(T,S) \
+  (sizeof (S) == (T)->len && !memcmp ((T)->ptr + 1, (S), (T)->len - 1))
 
 /* statement info */
-typedef enum Vis
+enum Vis
 {
   V_dot = 0,  /* random pseudo */
   V_var = 1,  /* var decl/defn */
@@ -210,22 +212,18 @@ typedef enum Vis
   V_weak = 0x10,   /* weakly globalize */
   V_no_eol = 0x20, /* no end of line */
   V_prefix_comment = 0x40 /* prefixed comment */
-} Vis;
+};
 
-typedef struct Stmt
+struct Stmt
 {
   struct Stmt *next;
   Token *tokens;
   Token *tokens_end;
-  symbol *sym;
   unsigned char vis;
-} Stmt;
+};
 
 #define append_stmt(V, S) ((S)->next = *(V), *(V) = (S))
 
-static Stmt *decls;
-static Stmt *fns;
-
 /* Read the whole input file.  It will be NUL terminated (but
    remember, there could be a NUL in the file itself.  */
 
@@ -267,7 +265,7 @@ read_file (FILE *stream)
    If we read a comment, append it to the comments block. */
 
 static Token *
-tokenize (const char *ptr)
+tokenize (const char *ptr, std::ostream &error_stream)
 {
   unsigned alloc = 1000;
   unsigned num = 0;
@@ -289,7 +287,11 @@ tokenize (const char *ptr)
 	goto block_comment;
       c = (unsigned char)*ptr++;
       if (c > 127)
-	fatal_error ("non-ascii character encountered: 0x%x", c);
+	{
+	  error_stream << "non-ascii character encountered: " << std::hex << std::showbase << (int) c;
+	  XDELETEVEC (toks);
+	  return NULL;
+	}
       switch (kind = c)
 	{
 	default:
@@ -438,10 +440,10 @@ tokenize (const char *ptr)
 /* Write an encoded token. */
 
 static void
-write_token (FILE *out, Token const *tok)
+write_token (std::ostream &out_stream, Token const *tok)
 {
   if (tok->space)
-    fputc (' ', out);
+    out_stream << ' ';
 
   switch (tok->kind)
     {
@@ -450,39 +452,99 @@ write_token (FILE *out, Token const *tok)
 	const char *c = tok->ptr + 1;
 	size_t len = tok->len - 2;
 
-	fputs ("\"", out);
+	out_stream << '"';
 	while (len)
 	  {
 	    const char *bs = (const char *)memchr (c, '\\', len);
 	    size_t l = bs ? bs - c : len;
 
-	    fprintf (out, "%.*s", (int)l, c);
+	    out_stream.write (c, l);
 	    len -= l;
 	    c += l;
 	    if (bs)
 	      {
-		fputs ("\\\\", out);
+		out_stream << "\\\\";
 		len--, c++;
 	      }
 	  }
-	fputs ("\"", out);
+	out_stream << '"';
       }
       break;
 
     default:
-      /* All other tokens shouldn't have anything magic in them */
-      fprintf (out, "%.*s", tok->len, tok->ptr);
+      /* All other tokens shouldn't have anything magic in them.  */
+      out_stream.write (tok->ptr, tok->len);
       break;
     }
 
   if (tok->end)
-    fputs ("\n", out);
+    out_stream << '\n';
+}
+
+/* The preamble '.target' directive's argument.  */
+static char *preamble_target_arg;
+
+/* Minimalistic verification of the preamble as generated by GCC.  */
+
+static bool
+verify_preamble (const Token *tok, std::ostream &error_stream)
+{
+  while (tok->kind == K_comment)
+    tok++;
+  if (tok->kind == K_dotted && is_keyword (tok, "version"))
+    tok++;
+  else
+    {
+      error_stream << "missing .version directive";
+      return false;
+    }
+  /* 'ptxas' doesn't seem to allow comments or line breaks here, but it's
+     not documented.  */
+  while (tok->kind == K_comment)
+    tok++;
+  if (tok->kind == K_number)
+    tok++;
+  else
+    {
+      error_stream << "malformed .version directive";
+      return false;
+    }
+  while (tok->kind == K_comment)
+    tok++;
+  if (tok->kind == K_dotted && is_keyword (tok, "target"))
+    tok++;
+  else
+    {
+      error_stream << "missing .target directive";
+      return false;
+    }
+  while (tok->kind == K_comment)
+    tok++;
+  if (tok->kind == K_symbol)
+    {
+      assert (!preamble_target_arg);
+      preamble_target_arg = xstrndup (tok->ptr, tok->len);
+      tok++;
+    }
+  else
+    {
+      error_stream << "malformed .target directive";
+      return false;
+    }
+  /* PTX allows here a "comma separated list of target specifiers", but GCC
+     doesn't generate that, and we don't support that.  */
+  if (tok->kind == ',')
+    {
+      error_stream << "unsupported list in .target directive";
+      return false;
+    }
+  return true;
 }
 
 static std::list<void *> heaps;
 
 static Stmt *
-alloc_stmt (unsigned vis, Token *tokens, Token *tokens_end, symbol *sym)
+alloc_stmt (unsigned vis, Token *tokens, Token *tokens_end)
 {
   static unsigned alloc = 0;
   static Stmt *heap = 0;
@@ -502,7 +564,6 @@ alloc_stmt (unsigned vis, Token *tokens, Token *tokens_end, symbol *sym)
   stmt->vis = vis;
   stmt->tokens = tokens;
   stmt->tokens_end = tokens_end;
-  stmt->sym = sym;
 
   return stmt;
 }
@@ -525,28 +586,26 @@ rev_stmts (Stmt *stmt)
 }
 
 static void
-write_stmt (FILE *out, const Stmt *stmt)
+write_stmt (std::ostream &out_stream, const Stmt *stmt)
 {
   for (Token *tok = stmt->tokens; tok != stmt->tokens_end; ++tok)
     {
       if ((stmt->vis & V_mask) == V_comment)
-	fprintf (out, "//");
-      write_token (out, tok);
-      if ((stmt->vis & V_mask) == V_pred)
-	fputc (' ', out);
+	out_stream << "//";
+      write_token (out_stream, tok);
     }
+  if ((stmt->vis & V_mask) == V_pred)
+    /* Space after the guard predicate.  */
+    out_stream << ' ';
 }
 
 static void
-write_stmts (FILE *out, const Stmt *stmts)
+write_stmts (std::ostream &out_stream, const Stmt *stmts)
 {
   for (; stmts; stmts = stmts->next)
-    write_stmt (out, stmts);
+    write_stmt (out_stream, stmts);
 }
 
-#define is_keyword(T,S) \
-  (sizeof (S) == (T)->len && !memcmp ((T)->ptr + 1, (S), (T)->len - 1))
-
 /* Parse a line until the end, regardless of semicolons.  */
 
 static Token *
@@ -558,14 +617,14 @@ parse_line_nosemi (unsigned vis, Token *tok, Stmt **append_where)
     tok++;
   while (!tok[-1].end);
 
-  Stmt *stmt = alloc_stmt (vis, start, tok, 0);
+  Stmt *stmt = alloc_stmt (vis, start, tok);
   append_stmt (append_where, stmt);
 
   return tok;
 }
 
 static Token *
-parse_insn (Token *tok)
+parse_insn (Stmt *&decls, Stmt *&fns, Token *tok)
 {
   unsigned depth = 0;
 
@@ -593,7 +652,7 @@ parse_insn (Token *tok)
 	case K_comment:
 	  while (tok->kind == K_comment)
 	    tok++;
-	  stmt = alloc_stmt (V_comment, start, tok, 0);
+	  stmt = alloc_stmt (V_comment, start, tok);
 	  append_stmt (&fns, stmt);
 	  continue;
 
@@ -611,9 +670,14 @@ parse_insn (Token *tok)
 	  break;
 
 	case '@':
-	  tok->space = 0;
 	  if (tok->kind == '!')
-	    tok++;
+	    {
+	      /* No space before '!'.  */
+	      tok->space = 0;
+	      tok++;
+	    }
+	  /* No space before the predicate variable.  */
+	  tok->space = 0;
 	  tok++;
 	  s = V_pred;
 	  break;
@@ -628,13 +692,13 @@ parse_insn (Token *tok)
 	  break;
 	}
 
-      stmt = alloc_stmt (s, start, tok, 0);
+      stmt = alloc_stmt (s, start, tok);
       append_stmt (&fns, stmt);
 
       if (!tok[-1].end && tok[0].kind == K_comment)
 	{
 	  stmt->vis |= V_no_eol;
-	  stmt = alloc_stmt (V_comment, tok, tok + 1, 0);
+	  stmt = alloc_stmt (V_comment, tok, tok + 1);
 	  append_stmt (&fns, stmt);
 	  tok++;
 	}
@@ -645,7 +709,7 @@ parse_insn (Token *tok)
 }
 
 static Token *
-parse_init (htab_t symbol_table, Token *tok, symbol *sym)
+parse_init (htab_t symbol_table, std::vector<symbol *> &symbols, Token *tok, symbol *sym)
 {
   for (;;)
     {
@@ -657,7 +721,7 @@ parse_init (htab_t symbol_table, Token *tok, symbol *sym)
 	{
 	  while (tok->kind == K_comment)
 	    tok++;
-	  stmt = alloc_stmt (V_comment, start, tok, 0);
+	  stmt = alloc_stmt (V_comment, start, tok);
 	  append_stmt (&sym->stmts, stmt);
 	  start = tok;
 	}
@@ -670,17 +734,17 @@ parse_init (htab_t symbol_table, Token *tok, symbol *sym)
 	if (tok->kind == K_symbol || tok->kind == K_ident)
 	  def_tok = tok;
       if (def_tok)
-	sym->deps.push_back (symbol_hash_lookup (symbol_table,
+	sym->deps.push_back (symbol_hash_lookup (symbol_table, symbols,
 						 xstrndup (def_tok->ptr,
 							   def_tok->len)));
       tok[1].space = 0;
       int end = tok++->kind == ';';
-      stmt = alloc_stmt (V_insn, start, tok, 0);
+      stmt = alloc_stmt (V_insn, start, tok);
       append_stmt (&sym->stmts, stmt);
       if (!tok[-1].end && tok->kind == K_comment)
 	{
 	  stmt->vis |= V_no_eol;
-	  stmt = alloc_stmt (V_comment, tok, tok + 1, 0);
+	  stmt = alloc_stmt (V_comment, tok, tok + 1);
 	  append_stmt (&sym->stmts, stmt);
 	  tok++;
 	}
@@ -691,7 +755,7 @@ parse_init (htab_t symbol_table, Token *tok, symbol *sym)
 }
 
 static Token *
-parse_file (htab_t symbol_table, Token *tok)
+parse_file (Stmt *&decls, htab_t symbol_table, std::vector<symbol *> &symbols, Stmt *&fns, Token *tok, std::ostream &error_stream)
 {
   Stmt *comment = 0;
   bool is_map_directive = false;
@@ -715,7 +779,7 @@ parse_file (htab_t symbol_table, Token *tok)
 	}
       if (start != tok)
 	{
-	  comment = alloc_stmt (V_comment, start, tok, 0);
+	  comment = alloc_stmt (V_comment, start, tok);
 	  comment->vis |= V_prefix_comment;
 	}
     }
@@ -742,7 +806,6 @@ parse_file (htab_t symbol_table, Token *tok)
       else
 	{
 	  unsigned vis = 0;
-	  symbol *def = 0;
 	  bool is_decl = false;
 	  Token *start, *def_token = 0;
 
@@ -774,9 +837,6 @@ parse_file (htab_t symbol_table, Token *tok)
 	      if (tok->kind == K_symbol || tok->kind == K_ident)
 		def_token = tok;
 	    }
-	  if (def_token)
-	    def = symbol_hash_lookup (symbol_table,
-				      xstrndup (def_token->ptr, def_token->len));
 
 	  if (!tok->kind)
 	    {
@@ -788,14 +848,14 @@ parse_file (htab_t symbol_table, Token *tok)
 		   || tok->kind == K_comment)
 	    {
 	      /* function defn */
-	      Stmt *stmt = alloc_stmt (vis, start, tok, def);
+	      Stmt *stmt = alloc_stmt (vis, start, tok);
 	      if (comment)
 		{
 		  append_stmt (&fns, comment);
 		  stmt->vis |= V_prefix_comment;
 		}
 	      append_stmt (&fns, stmt);
-	      tok = parse_insn (tok);
+	      tok = parse_insn (decls, fns, tok);
 	    }
 	  else
 	    {
@@ -804,17 +864,25 @@ parse_file (htab_t symbol_table, Token *tok)
 	      tok++->end = 1;
 	      if ((vis & V_mask) == V_var && !is_decl)
 		{
-		  if (def == NULL)
+		  if (!def_token)
 		    {
 		      const char *eol = strchr (start->ptr, '\n');
 		      const char *line
 			= ((eol == NULL)
 			   ? start->ptr
 			   : strndup (start->ptr, eol - start->ptr));
-		      fatal_error ("expected identifier in line '%s'", line);
+		      {
+			error_stream << "expected identifier in line '" << line << "'";
+			if (line != start->ptr)
+			  free ((void *) line);
+			return NULL;
+		      }
 		    }
+
 		  /* variable */
-		  Stmt *stmt = alloc_stmt (vis, start, tok, def);
+		  symbol *def = symbol_hash_lookup (symbol_table, symbols,
+						    xstrndup (def_token->ptr, def_token->len));
+		  Stmt *stmt = alloc_stmt (vis, start, tok);
 		  if (comment)
 		    {
 		      append_stmt (&def->stmts, comment);
@@ -822,12 +890,12 @@ parse_file (htab_t symbol_table, Token *tok)
 		    }
 		  append_stmt (&def->stmts, stmt);
 		  if (assign)
-		    tok = parse_init (symbol_table, tok, def);
+		    tok = parse_init (symbol_table, symbols, tok, def);
 		}
 	      else
 		{
 		  /* declaration */
-		  Stmt *stmt = alloc_stmt (vis, start, tok, 0);
+		  Stmt *stmt = alloc_stmt (vis, start, tok);
 		  if (comment)
 		    {
 		      append_stmt (&decls, comment);
@@ -852,37 +920,36 @@ parse_file (htab_t symbol_table, Token *tok)
   return tok;
 }
 
-static void
-output_symbol (FILE *out, symbol *e)
+static bool
+symbol_order_deps (std::vector<const symbol *> &symbols_out, symbol *e, std::ostream &error_stream)
 {
   if (e->emitted)
-    return;
+    return true;
   if (e->pending)
-    fatal_error ("circular reference in variable initializers");
+    {
+      error_stream << "circular reference in variable initializers";
+      return false;
+    }
   e->pending = true;
   std::list<symbol *>::iterator i;
   for (i = e->deps.begin (); i != e->deps.end (); i++)
-    output_symbol (out, *i);
+    if (!symbol_order_deps (symbols_out, *i, error_stream))
+      return false;
   e->pending = false;
-  write_stmts (out, rev_stmts (e->stmts));
+  symbols_out.push_back (e);
   e->emitted = true;
-}
-
-static int
-traverse (void **slot, void *data)
-{
-  symbol *e = *(symbol **)slot;
-  output_symbol ((FILE *)data, e);
-  return 1;
+  return true;
 }
 
 static void
-process (FILE *in, FILE *out, int *verify, const char *inname)
+process (FILE *in, std::ostream &out_stream, int *verify, const char *inname)
 {
+  std::ostringstream error_stream;
+
   const char *input = read_file (in);
 
   /* As expected by GCC, handle an empty input file specially.  See
-     <https://github.com/MentorEmbedded/nvptx-tools/pull/26> "[nvptx-as] Allow
+     <https://github.com/SourceryTools/nvptx-tools/pull/26> "[nvptx-as] Allow
      empty input file" for reference.  */
   if (*input == '\0')
     {
@@ -896,7 +963,9 @@ process (FILE *in, FILE *out, int *verify, const char *inname)
       return;
     }
 
-  Token *tok = tokenize (input);
+  Token *tok = tokenize (input, error_stream);
+  if (!tok)
+    fatal_error (error_stream.str ());
   Token *tok_to_free = tok;
 
   /* Do minimalistic verification, so that we reliably reject (certain classes
@@ -904,60 +973,51 @@ process (FILE *in, FILE *out, int *verify, const char *inname)
      verify the whole output file.)  */
   if (*verify != 0)
     {
-      /* Verify the preamble as generated by GCC.  */
-      size_t i = 0;
-      while (tok[i].kind == K_comment)
-	i++;
-      if (tok[i].kind == K_dotted && is_keyword (&tok[i], "version"))
-	i++;
-      else
-	fatal_error ("missing .version directive at start of file '%s'",
-		     inname);
-      /* 'ptxas' doesn't seem to allow comments or line breaks here, but it's
-	 not documented.  */
-      while (tok[i].kind == K_comment)
-	i++;
-      if (tok[i].kind == K_number)
-	i++;
-      else
-	fatal_error ("malformed .version directive at start of file '%s'",
-		     inname);
-      while (tok[i].kind == K_comment)
-	i++;
-      if (tok[i].kind == K_dotted && is_keyword (&tok[i], "target"))
-	i++;
-      else
-	fatal_error ("missing .target directive at start of file '%s'",
-		     inname);
-      while (tok[i].kind == K_comment)
-	i++;
-      if (tok[i].kind == K_symbol)
+      if (!verify_preamble (tok, error_stream))
 	{
-	  assert (!preamble_target_arg);
-	  preamble_target_arg = xstrndup (tok[i].ptr, tok[i].len);
-	  i++;
+	  error_stream << " at start of file '" << inname << "'";
+	  fatal_error (error_stream.str ());
 	}
-      else
-	fatal_error ("malformed .target directive at start of file '%s'",
-		     inname);
-      /* PTX allows here a "comma separated list of target specifiers", but GCC
-	 doesn't generate that, and we don't support that.  */
-      if (tok[i].kind == ',')
-	fatal_error ("unsupported list in .target directive at start of file '%s'",
-		     inname);
     }
 
-  htab_t symbol_table
-    = htab_create (500, hash_string_hash, hash_string_eq, symbol_hash_free);
+  Stmt *decls = NULL;
+  /* Initial symbol table size.  */
+  const size_t n_symbols_init = 500;
+  htab_t symbol_table = htab_create (n_symbols_init, hash_string_hash, hash_string_eq, symbol_hash_free);
+  /* Symbols, in order of appearance.  */
+  std::vector<symbol *> symbols;
+  symbols.reserve (n_symbols_init);
+  Stmt *fns = NULL;
 
   do
-    tok = parse_file (symbol_table, tok);
+    {
+      tok = parse_file (decls, symbol_table, symbols, fns, tok, error_stream);
+      if (!tok)
+	fatal_error (error_stream.str ());
+    }
   while (tok->kind);
 
-  write_stmts (out, rev_stmts (decls));
-  htab_traverse (symbol_table, traverse, (void *)out);
-  write_stmts (out, rev_stmts (fns));
+  /* Actual number of symbols.  */
+  const size_t n_symbols = htab_elements (symbol_table);
+  /* All symbols handled.  */
+  assert (symbols.size () == n_symbols);
 
+  /* Symbols, in order of output.  */
+  std::vector<const symbol *> symbols_out;
+  symbols_out.reserve (n_symbols);
+  for (symbol *e : symbols)
+    if (!symbol_order_deps (symbols_out, e, error_stream))
+      fatal_error (error_stream.str ());
+  /* All symbols emitted.  */
+  assert (symbols_out.size () == n_symbols);
+
+  write_stmts (out_stream, rev_stmts (decls));
+  for (const symbol *e : symbols_out)
+    write_stmts (out_stream, rev_stmts (e->stmts));
+  write_stmts (out_stream, rev_stmts (fns));
+
+  symbols_out.clear ();
+  symbols.clear ();
   htab_delete (symbol_table);
 
   while (!heaps.empty ())
@@ -980,17 +1040,23 @@ collect_wait (const char *prog, struct pex_obj *pex)
   int status;
 
   if (!pex_get_status (pex, 1, &status))
-    fatal_error ("can't get program status: %m");
+    {
+      std::ostringstream error_stream;
+      error_stream << "can't get program status: " << strerror (errno);
+      fatal_error (error_stream.str ());
+    }
   pex_free (pex);
 
   if (status)
     {
       if (WIFSIGNALED (status))
 	{
+	  std::ostringstream error_stream;
 	  int sig = WTERMSIG (status);
-	  fatal_error ("%s terminated with signal %d [%s]%s",
-		       prog, sig, strsignal(sig),
-		       WCOREDUMP(status) ? ", core dumped" : "");
+	  error_stream << prog << " terminated with signal " << sig << " [" << strsignal (sig) << "]";
+	  if (WCOREDUMP (status))
+	    error_stream << ", core dumped";
+	  fatal_error (error_stream.str ());
 	}
 
       if (WIFEXITED (status))
@@ -1005,7 +1071,9 @@ do_wait (const char *prog, struct pex_obj *pex)
   int ret = collect_wait (prog, pex);
   if (ret != 0)
     {
-      fatal_error ("%s returned %d exit status", prog, ret);
+      std::ostringstream error_stream;
+      error_stream << prog << " returned " << ret << " exit status";
+      fatal_error (error_stream.str ());
     }
 }
 
@@ -1019,16 +1087,20 @@ fork_execute (const char *prog, const char *const *argv)
       for (const char *const *arg = argv; *arg; ++arg)
 	{
 	  if (**arg == '\0')
-	    fprintf (stderr, " ''");
+	    std::cerr << " ''";
 	  else
-	    fprintf (stderr, " %s", *arg);
+	    std::cerr << " " << *arg;
 	}
-      fprintf (stderr, "\n");
+      std::cerr << "\n";
     }
 
   struct pex_obj *pex = pex_init (0, "nvptx-as", NULL);
   if (pex == NULL)
-    fatal_error ("pex_init failed: %m");
+    {
+      std::ostringstream error_stream;
+      error_stream << "pex_init failed: " << strerror (errno);
+      fatal_error (error_stream.str ());
+    }
 
   int err;
   const char *errmsg;
@@ -1038,13 +1110,12 @@ fork_execute (const char *prog, const char *const *argv)
 		    NULL, NULL, &err);
   if (errmsg != NULL)
     {
+      std::ostringstream error_stream;
+      error_stream << "error trying to exec '" << argv[0] << "': ";
+      error_stream << errmsg;
       if (err != 0)
-	{
-	  errno = err;
-	  fatal_error ("error trying to exec '%s': %s: %m", argv[0], errmsg);
-	}
-      else
-	fatal_error ("error trying to exec '%s': %s", argv[0], errmsg);
+	error_stream << ": " << strerror (err);
+      fatal_error (error_stream.str ());
     }
   do_wait (prog, pex);
 }
@@ -1123,9 +1194,9 @@ program_available (const char *progname)
 }
 
 ATTRIBUTE_NORETURN static void
-usage (FILE *stream, int status)
+usage (std::ostream &out_stream, int status)
 {
-  fprintf (stream, "\
+  out_stream << "\
 Usage: nvptx-none-as [option...] [asmfile]\n\
 Options:\n\
   -m TARGET             Override target architecture used for ptxas\n\
@@ -1137,8 +1208,7 @@ Options:\n\
   --help                Print this help and exit\n\
   --version             Print version number and exit\n\
 \n\
-Report bugs to %s.\n",
-	   REPORT_BUGS_TO);
+Report bugs to " << REPORT_BUGS_TO << ".\n";
   exit (status);
 }
 
@@ -1157,7 +1227,7 @@ main (int argc, char **argv)
 {
   FILE *in = stdin;
   const char *inname = "{standard input}";
-  FILE *out = stdout;
+  std::ostream *out_stream = &std::cout;
   int verify = -1;
   const char *target_arg_force = NULL;
 
@@ -1179,7 +1249,7 @@ main (int argc, char **argv)
 	case 'o':
 	  if (outname != NULL)
 	    {
-	      fprintf (stderr, "multiple output files specified\n");
+	      std::cerr << "multiple output files specified\n";
 	      exit (1);
 	    }
 	  outname = optarg;
@@ -1191,20 +1261,19 @@ main (int argc, char **argv)
 	  /* Ignore include paths.  */
 	  break;
 	case 'h':
-	  usage (stdout, 0);
+	  usage (std::cout, 0);
 	  break;
 	case 'V':
-	  printf ("\
-nvptx-none-as %s%s\n\
-Copyright %s Mentor Graphics\n\
+	  std::cout << "\
+nvptx-none-as " << PKGVERSION << NVPTX_TOOLS_VERSION << "\n\
+Copyright (C) 2024 The nvptx-tools Developers\n\
 License GPLv3+: GNU GPL version 3 or later <http://gnu.org/licenses/gpl.html>\n\
 This program is free software; you may redistribute it under the terms of\n\
 the GNU General Public License version 3 or later.\n\
-This program has absolutely no warranty.\n",
-		  PKGVERSION, NVPTX_TOOLS_VERSION, "2015");
+This program has absolutely no warranty.\n";
 	  exit (0);
 	default:
-	  usage (stderr, 1);
+	  usage (std::cerr, 1);
 	  break;
 	}
     }
@@ -1213,9 +1282,13 @@ This program has absolutely no warranty.\n",
     fatal_error ("too many input files specified");
 
   if (outname)
-    out = fopen (outname, "w");
-  if (!out)
-    fatal_error ("cannot open '%s'", outname);
+    out_stream = new std::ofstream (outname);
+  if (out_stream->fail ())
+    {
+      std::ostringstream error_stream;
+      error_stream << "cannot open '" << outname << "'";
+      fatal_error (error_stream.str ());
+    }
 
   if (argc > optind)
     {
@@ -1225,17 +1298,17 @@ This program has absolutely no warranty.\n",
   if (!in)
     fatal_error ("cannot open input ptx file");
 
-  process (in, out, &verify, inname);
+  process (in, *out_stream, &verify, inname);
 
   if (in != stdin)
     {
       fclose (in);
       in = NULL;
     }
-  if (out != stdout)
+  if (out_stream != &std::cout)
     {
-      fclose (out);
-      out = NULL;
+      delete out_stream;
+      out_stream = NULL;
     }
 
   if (outname == NULL)
@@ -1290,10 +1363,10 @@ This program has absolutely no warranty.\n",
 		 'sm_50', which is the baseline supported by all current CUDA
 		 versions down to CUDA 6.5, at least.  */
 	      if (verbose)
-		fprintf (stderr, "Verifying %s code", target_arg);
+		std::cerr << "Verifying " << target_arg << " code";
 	      target_arg = "sm_50";
 	      if (verbose)
-		fprintf (stderr, " with %s code generation.\n", target_arg);
+		std::cerr << " with " << target_arg << " code generation.\n";
 	    }
 	}
 
@@ -1313,7 +1386,7 @@ This program has absolutely no warranty.\n",
   else if (verify < 0)
     {
       if (verbose)
-	fprintf (stderr, "'ptxas' not available.\n");
+	std::cerr << "'ptxas' not available.\n";
     }
 
   free (preamble_target_arg);
