@@ -1,7 +1,8 @@
 /* List symbols.
    Copyright (C) 1991-2022 Free Software Foundation, Inc.
    Copyright (C) 2014, 2015 Mentor Graphics
-   Copyright (C) 2022 Siemens
+   Copyright (C) 2022, 2023 Siemens
+   Copyright (C) 2024 BayLibre
 
    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
@@ -54,7 +55,7 @@
 
 #include "version.h"
 
-ATTRIBUTE_NORETURN static void usage (FILE *, int);
+ATTRIBUTE_NORETURN static void usage (std::ostream &, int);
 
 static const char *print_format_string = NULL;
 
@@ -68,12 +69,12 @@ static int filename_per_file = 0;	/* Once per file, on its own line.  */
 static int print_width = 0;
 static int print_radix = 16;
 
-typedef struct symbol_hash_entry
+struct symbol_hash_entry
 {
   /* The name of the symbol.  */
   const char *key;
   char type;
-} symbol;
+};
 
 static void
 symbol_hash_free (void *elt)
@@ -130,23 +131,30 @@ symbol_hash_lookup (htab_t symbol_table, char *string, int create)
   return (struct symbol_hash_entry *) *e;
 }
 
+/* Process GCC/nvptx-generated linker markers.
+
+   Capture in 'symbol_table_global'/'symbol_table_local' the type of symbols in
+   'ptx'.  */
+
 static const char *
 process_refs_defs (htab_t symbol_table_global, htab_t symbol_table_local, const char *ptx)
 {
   /* There is no actual symbol table in GCC/nvptx.  We only look for the
      GCC/nvptx-generated linker markers, not for actual PTX-level symbols.
+     (We do look for PTX '.weak', though.)
      We therefore only present the information conveyed by those markers.  */
   while (*ptx != '\0')
     {
+      const char *ptx_begin = ptx;
       bool global_p;
       if ((global_p = (strncmp (ptx, "\n// BEGIN GLOBAL ", 17) == 0))
 	  || (strncmp (ptx, "\n// BEGIN ", 10) == 0))
 	{
-	  char type = '\0';
 	  if (global_p)
 	    ptx += 17;
 	  else
 	    ptx += 10;
+	  char type;
 	  if (strncmp (ptx, "VAR DEF: ", 9) == 0)
 	    {
 	      type = global_p ? 'D' : 'd';
@@ -157,7 +165,7 @@ process_refs_defs (htab_t symbol_table_global, htab_t symbol_table_local, const
 	      type = global_p ? 'T' : 't';
 	      ptx += 14;
 	    }
-	  if (strncmp (ptx, "VAR DECL: ", 10) == 0)
+	  else if (strncmp (ptx, "VAR DECL: ", 10) == 0)
 	    {
 	      type = 'U';
 	      ptx += 10;
@@ -167,38 +175,103 @@ process_refs_defs (htab_t symbol_table_global, htab_t symbol_table_local, const
 	      type = 'U';
 	      ptx += 15;
 	    }
+	  else
+	    /* Unknown marker: ignore, and hope for the best.  */
+	    type = '\0';
+
+	  const char *ptx_sym_name_begin = ptx;
+	  const char *ptx_lf = strchr (ptx, '\n');
+	  if (!ptx_lf)
+	    {
+	      assert (ptx_begin[0] == '\n');
+	      std::cerr << "error, truncated marker line:" << ptx_begin << "\n";
+	      return NULL;
+	    }
+	  const char *ptx_sym_name_end = ptx_lf;
+	  ptx = ptx_lf;
+
 	  if (type == '\0')
 	    continue;
-	  const char *end = strchr (ptx, '\n');
-	  if (end == 0)
-	    end = ptx + strlen (ptx);
 
-	  char *sym = xstrndup (ptx, end - ptx);
+	  /* GCC/nvptx currently emits the PTX '.weak' linking directive
+	     without special annotation in the corresponding marker line.
+	     Per 'as', '.weak ' follows directly after the marker line.  */
+	  bool ptx_weak_p = strncmp (ptx, "\n.weak ", 7) == 0;
+	  if (ptx_weak_p)
+	    ptx += 7;
+
+	  char *sym_name = xstrndup (ptx_sym_name_begin, ptx_sym_name_end - ptx_sym_name_begin);
+
+	  if (ptx_weak_p)
+	    {
+	      if (!global_p)
+		{
+		  const char *type_s;
+		  if (type == 'd')
+		    type_s = "VAR DEF";
+		  else if (type == 't')
+		    type_s = "FUNCTION DEF";
+		  else if (type == 'U')
+		    type_s = "DECL";
+		  else
+		    abort ();
+		  std::cerr << "error, non-GLOBAL " << type_s << " is \"weak\": '" << sym_name << "'\n";
+
+		  free (sym_name);
+
+		  return NULL;
+		}
+	      else if (type == 'D')
+		type = 'V';
+	      else if (type == 'T')
+		type = 'W';
+	      else if (type == 'U')
+		type = 'w';
+	      else
+		abort ();
+	    }
+
 	  htab_t symbol_table
 	    = global_p ? symbol_table_global : symbol_table_local;
 	  struct symbol_hash_entry *e
-	    = symbol_hash_lookup (symbol_table, sym, 1);
+	    = symbol_hash_lookup (symbol_table, sym_name, 1);
 
 	  if (e->type == '\0')
 	    {
 	      /* First time we're seeing this.  */
 	      e->type = type;
 	    }
-	  else if (e->type == 'U')
+	  else if (e->type == 'U'
+		   || e->type == 'w')
 	    {
 	      /* We've seen this before, but not a definition.  */
 	      e->type = type;
 	    }
-	  else if (type == 'U')
+	  else if (type == 'U'
+		   || type == 'w')
 	    {
 	      /* We've already seen a definition.  */
 	    }
+	  else if ((type == 'V' && e->type == 'D')
+		   || (type == 'W' && e->type == 'T'))
+	    {
+	      /* This "weak" definition doesn't affect the former "strong" one.  */
+	    }
+	  else if ((type == 'D' && e->type == 'V')
+		   || (type == 'T' && e->type == 'W'))
+	    {
+	      /* This "strong" definition overrides the former "weak" one.  */
+	      e->type = type;
+	    }
 	  else
 	    {
 	      /* We've already seen a definition; verify it's the same now.  */
 	      assert (e->type == type);
 	    }
+
+	  continue;
 	}
+
       ptx++;
     }
   /* Callers may use this return value to detect NUL-separated parts.  */
@@ -230,7 +303,7 @@ set_output_format (const char *f)
       break;
     default:
       std::cerr << f << ": invalid output format\n";
-      usage (stderr, 1);
+      usage (std::cerr, 1);
     }
 }
 
@@ -285,12 +358,15 @@ numeric_forward (const void *P_x, const void *P_y)
   const symbol_hash_entry *xe = *(const symbol_hash_entry **) P_x;
   const symbol_hash_entry *ye = *(const symbol_hash_entry **) P_y;
 
-  if (xe->type == 'U')
+  if (xe->type == 'U'
+      || xe->type == 'w')
     {
-      if (ye->type != 'U')
+      if (!(ye->type == 'U'
+	    || ye->type == 'w'))
 	return -1;
     }
-  else if (ye->type == 'U')
+  else if (ye->type == 'U'
+	   || ye->type == 'w')
     return 1;
 #if 0
   /* There are no symbol values in GCC/nvptx.  */
@@ -336,7 +412,8 @@ print_value (void *val)
 static void
 print_symbol_info_bsd (symbol_hash_entry *sym)
 {
-  if (sym->type == 'U')
+  if (sym->type == 'U'
+      || sym->type == 'w')
     {
       if (print_width == 64)
 	printf ("        ");
@@ -477,9 +554,9 @@ print_object_filename_bsd (const char *filename)
 }
 
 ATTRIBUTE_NORETURN static void
-usage (FILE *stream, int status)
+usage (std::ostream &out_stream, int status)
 {
-  fprintf (stream, "\
+  out_stream << "\
 Usage: nvptx-none-nm [option...] [files...]\n\
 Options:\n\
   -B                    Same as --format=bsd\n\
@@ -493,8 +570,7 @@ Options:\n\
 \n\
 If no files are given, `a.out' is defaulted.\n\
 \n\
-Report bugs to %s.\n",
-	   REPORT_BUGS_TO);
+Report bugs to " << REPORT_BUGS_TO << ".\n";
   exit (status);
 }
 
@@ -546,22 +622,21 @@ main (int argc, char **argv)
 	  reverse_sort = 1;
 	  break;
 	case 'h':
-	  usage (stdout, 0);
+	  usage (std::cout, 0);
 	  break;
 	case 'V':
-	  printf ("\
-nvptx-none-nm %s%s\n\
-Copyright %s Free Software Foundation, Inc., Mentor Graphics, Siemens\n\
+	  std::cout << "\
+nvptx-none-nm " << PKGVERSION << NVPTX_TOOLS_VERSION << "\n\
+Copyright (C) 2024 The nvptx-tools Developers\n\
 License GPLv3+: GNU GPL version 3 or later <http://gnu.org/licenses/gpl.html>\n\
 This program is free software; you may redistribute it under the terms of\n\
 the GNU General Public License version 3 or later.\n\
-This program has absolutely no warranty.\n",
-		  PKGVERSION, NVPTX_TOOLS_VERSION, "2022");
+This program has absolutely no warranty.\n";
 	  exit (0);
 	case 0:		/* A long option that just sets a flag.  */
 	  break;
 	default:
-	  usage (stderr, 1);
+	  usage (std::cerr, 1);
 	  break;
 	}
     }
@@ -574,6 +649,7 @@ This program has absolutely no warranty.\n",
   if (inputfiles.size () > 1)
     filename_per_file = 1;
 
+  /* Scan 'inputfiles'.  */
   for (std::list<std::string>::const_iterator iterator = inputfiles.begin(), end = inputfiles.end();
        iterator != end;
        ++iterator)
@@ -604,8 +680,9 @@ This program has absolutely no warranty.\n",
       set_print_width (buf);
 
       std::list<htab_t> symbol_tables;
-      htab_t symbol_table_global
-	= htab_create (500, hash_string_hash, hash_string_eq, symbol_hash_free);
+      /* Initial symbol table size.  */
+      const size_t n_symbols_init = 500;
+      htab_t symbol_table_global = htab_create (n_symbols_init, hash_string_hash, hash_string_eq, symbol_hash_free);
       symbol_tables.push_back (symbol_table_global);
       const char *buf_ = buf;
       /* Per nvptx-tools 'ld', there are several NUL-separated parts.  These we
@@ -613,21 +690,30 @@ This program has absolutely no warranty.\n",
 	 symbols.  */
       do
 	{
-	  htab_t symbol_table_local
-	    = htab_create (500, hash_string_hash, hash_string_eq, symbol_hash_free);
+	  htab_t symbol_table_local = htab_create (n_symbols_init, hash_string_hash, hash_string_eq, symbol_hash_free);
 	  symbol_tables.push_back (symbol_table_local);
 	  buf_ = process_refs_defs (symbol_table_global, symbol_table_local, buf_);
+	  if (buf_ == NULL)
+	    {
+	      std::cerr << "while scanning '" << name << "'\n";
+	      break;
+	    }
 	}
       while (buf_ != &buf[len + 1]);
       delete[] buf;
-      print_object_filename_bsd (name.c_str ());
-      display_rel_file (symbol_tables, name);
+      if (buf_ != NULL)
+	{
+	  print_object_filename_bsd (name.c_str ());
+	  display_rel_file (symbol_tables, name);
+	}
       while (!symbol_tables.empty ())
 	{
 	  htab_t symbol_table = symbol_tables.front ();
 	  htab_delete (symbol_table);
 	  symbol_tables.pop_front ();
 	}
+      if (buf_ == NULL)
+	goto error_out;
     }
 
   free ((char *) print_format_string);
